//============================================================================
//  aPCI-8131 for PCI Linux Device Driver
//  apci8131.c
//  Author  Shoichiro Ito(ADTEK SYSTEM SCIENCE Co.,Ltd.)
//  Date    2001/02/28
//  Rev.    1.0.0
//============================================================================

#ifndef MODULE
#define MODULE
#endif // MODULE

#ifndef __KERNEL__
#define __KERNEL__
#endif // __KERNEL__

#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#	define MODVERSIONS
#endif

#ifdef MODVERSIONS
#	include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>		// used file oprations
#include <linux/pci.h>		// used pci functions
#include <asm/io.h>		// used inb, outb
#include <asm/uaccess.h>	// used verify_area
#include <linux/poll.h>		// used poll_wait
#include <asm/delay.h>		// used udelay

#include "apci8131.h"

/*
 * Split minors in two parts
 */
#define TYPE(dev)		(MINOR(dev) >> 4)	// high nibble
#define NUM(dev)		(MINOR(dev) & 0xf)	// low	nibble


#define APCI8131_NAME		"apci8131"

#define APCI8131_VENDOR_ID	0x136C		// ǥХID,٥ID פʤ
#define APCI8131_DEVICE_ID	0x8131		// ɥ饤ФϿǤʤΤ

#define APCI8131_INT_OCCURS	1		// ȯ
#define APCI8131_INT_READY	0		// ȯԤ

#ifdef APCI8131_DEBUG
#	define APCI8131_DELAY_CLOSE	1000 	// Ԥ
#else
#	define APCI8131_DELAY_CLOSE	200 	// Ԥ
#endif

#define PWARNING(fmt, args...)	printk(KERN_WARNING "[APCI8131]: " fmt, ## args)

#define APIC21_IFR_PORT		0x02		// Interrupt Flag Register
#define APIC21_ICR_PORT		0x03		// Interrupt Clear Register
#define APCI21_EICR_PORT	0x0E		// External Interrupt Control Register

#define APIC21_IFR_MSK		0x0F		// Interrupt Flag Mask
#define APIC21_ICR_ALL		0x0F		// Interrupt All Clear
#define APIC21_EICR_DIS		0x00		// External Interrupt Control Disable


/*----------------- Define Device Information ------------------------------*/

typedef struct _devinf
{
	unsigned int		apic_addr;	// APIC21 Base Address
	unsigned int		io_addr;	// I/O	  Base Address
	unsigned char		irq;		// Interrupt Request
	unsigned long		f_version;	// ե뼱̻
	struct wait_queue*	wq;		// Ե
	int			int_occurs;	// ȯ
	unsigned char		ifr;		// װ
} devinf, *pdevinf;

static devinf		devctl[APCI8131_MAX_DEVICES];	// ǥХ
static unsigned int	uapci8131major;			// Major Device Number

static int devmajor = 0;


/*----------------- File Operations ----------------------------------------*/

static unsigned int apci8131_poll(struct file* file, struct poll_table_struct* ptab);
static		int apci8131_ioctl(struct inode*, struct file*, unsigned int, unsigned long);
static		int apci8131_open(struct inode* inode, struct file* file);
static		int apci8131_close(struct inode* inode, struct file* file);

static struct file_operations apci8131_fops =
{
	NULL,		// loff_t	llseek(struct file*, loff_t, int)
	NULL,		// ssize_t	read(struct file*, char*, size_t, loff_t*)
	NULL,		// ssize_t	write(struct file*, const char*, size_t, loff_t*)
	NULL,		// int		readdir(struct file*, void*, filldir_t)
	apci8131_poll,	// unsigned int poll(struct file*, struct poll_table_struct*)
	apci8131_ioctl,	// int		ioctl(struct inode*, struct file*, unsigned int, unsigned long)
	NULL,		// int		mmap(struct file*, struct vm_area_struct*)
	apci8131_open,	// int		open(struct inode*, struct file*)
	NULL,		// int		flush(struct file*)
	apci8131_close,	// int		release(struct inode*, struct file*)
	NULL,		// int		fsync(struct file*, struct dentry*)
	NULL,		// int		fasync(int, struct file*, int)
	NULL,		// int		check_media_change(kdev_t dev)
	NULL,		// int		revalidate(kdev_t dev)
	NULL,		// int		lock(struct file*, int, struct file_lock*)
};


/*----------------- Define Static Functions --------------------------------*/

static int		init_driver_info(void);
static int		setup_driver(void);
static void		release_driver(void);
static int		get_device(void);
static unsigned int	get_device_base_addr(unsigned int base_addr);
static int		get_bsn(unsigned int io_addr);
static int		regist_char_driver(void);
static int		unregist_char_driver(void);
static int		get_device_state(int ref);

static int	setup_hw_isr(int ref);
static void	release_hw_isr(int ref);

static int	verify_arg(unsigned int cmd, unsigned long arg);
static int	apci8131_ioc_inp(struct file* file, unsigned long arg);
static int	apci8131_ioc_outp(struct file* file, unsigned long arg);
static int	apci8131_ioc_get_ifr(struct file* file, unsigned long arg);
static int	apci8131_ioc_get_icr(struct file* file, unsigned long arg);
static int	apci8131_ioc_set_icr(struct file* file, unsigned long arg);

static void	apci8131_isr(int irq, void* dev_id, struct pt_regs* regs);

static int	is_port(unsigned port);

static inline unsigned char  apic_get_ifr(int ref);
static inline void	     apic_set_icr(int ref, unsigned char icr);
static inline unsigned short apic_get_eicr(int ref);
static inline void	     apic_set_eicr(int ref, unsigned short eicr);


MODULE_PARM(devmajor, "i");	// ᥸㡼ֹѥ᡼


/*==========================================================================*/
/*  init_module - ɥ饤ФϿ					    */
/*==========================================================================*/
int init_module(void)
{
	int result;

	// PCI ǽ̵ͭȽ
	if (!pci_present())
		return -ENODEV;

	// ǥХɥ饤ѿν
	result = init_driver_info();
	if (result < 0)
		return result;

	// ǥХɥ饤ФΥåȥå
	result = setup_driver();
	if (result < 0)
		return result;

	return 0;
}

/*==========================================================================*/
/*  cleanup_module - ɥ饤ФϿä				    */
/*==========================================================================*/
void cleanup_module(void)
{
	// ǥХɥ饤Фβ
	release_driver();
}

/*--------------------------------------------------------------------------*/
/*  apci8131_poll - ǥХߤΥݡ			    */
/*--------------------------------------------------------------------------*/
static unsigned int apci8131_poll(struct file* file, struct poll_table_struct* ptab)
{
	int ref;

	PDEBUG("Poll\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ȯ
	if (devctl[ref].int_occurs == APCI8131_INT_OCCURS)
	{
		devctl[ref].int_occurs = APCI8131_INT_READY;	// ȯԤذܹ
		return POLLIN | POLLRDNORM;
	}

	poll_wait(file, &devctl[ref].wq, ptab);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_ioctl - ǥХͭ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg)
{
	int err = verify_arg(cmd, arg);
	if (err)
		return err;

	switch (cmd)
	{
		case APCI8131_IN_PORT:
			return apci8131_ioc_inp(file, arg);

		case APCI8131_OUT_PORT:
			return apci8131_ioc_outp(file, arg);

		case APCI8131_GET_IFR:
			return apci8131_ioc_get_ifr(file, arg);

		case APCI8131_GET_ICR:
			return apci8131_ioc_get_icr(file, arg);

		case APCI8131_SET_ICR:
			return apci8131_ioc_set_icr(file, arg);
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_open - ǥХΥץ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_open(struct inode* inode, struct file* file)
{
	int ref;
	int result;

	PDEBUG("Open !! %ld\n", file->f_version);

	// Ȥ
	ref = NUM(inode->i_rdev);

	PDEBUG("ref index=%d\n", ref);

	// ǥХξ
	result = get_device_state(ref);
	if (result < 0)
		return result;

	// ߤ򤹤
	result = setup_hw_isr(ref);
	if (result < 0)
		return result;

	// ߴϢѿ
	devctl[ref].wq = NULL;
	devctl[ref].int_occurs = APCI8131_INT_READY;

	// Ѿ֤ˤ
	devctl[ref].f_version = file->f_version;

	// ǡȰ¸
	file->private_data = (void*)ref;

	MOD_INC_USE_COUNT;		// ȥ+

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_close - ǥХΥ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_close(struct inode* inode, struct file* file)
{
	int ref;

	PDEBUG("Close !!\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// ѿʤɤΥ꡼󥢥å׽

	// ߶ػ
	apic_set_eicr(ref, APIC21_EICR_DIS);

	// ˳ߤäΤȤθ
	// ΥȤ
	udelay(APCI8131_DELAY_CLOSE);

	// ȯѿꥻå
	devctl[ref].int_occurs = APCI8131_INT_READY;

	// Ԥ󤬤ʤ
	while (waitqueue_active(&devctl[ref].wq))
		wake_up(&devctl[ref].wq);

	// ߤ
	release_hw_isr(ref);

	// ̤Ѿ֤ˤ
	devctl[ref].f_version = 0;

	MOD_DEC_USE_COUNT;		// ȥ-

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  init_driver_info - ǥХɥ饤дѿν			    */
/*  Input : none							    */
/*  Output: return - 0 					    */
/*--------------------------------------------------------------------------*/
static int init_driver_info(void)
{
	int i;

	for (i = 0; i < APCI8131_MAX_DEVICES; i++)
	{
		devctl[i].apic_addr = 0x0000;
		devctl[i].io_addr   = 0x0000;
		devctl[i].irq	    = 0;
		devctl[i].f_version = 0;
		devctl[i].ifr	    = 0;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  setup_driver - ǥХɥ饤ФΥåȥå			    */
/*  Input : none							    */
/*  Output: return - 0 ʳ åȥå׼				    */
/*		     0 åȥå					    */
/*--------------------------------------------------------------------------*/
static int setup_driver(void)
{
	int result;

	result = get_device();
	if (result < 0)
		return result;

	// character device driver registration
	result = regist_char_driver();
	if (result < 0)
		return result;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  release_driver -  ǥХɥ饤Фβ				    */
/*  Input : none							    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void release_driver(void)
{
	// character device driver unregistration
	unregist_char_driver();
}

/*--------------------------------------------------------------------------*/
/*  get_device - 󥹥ȡ뤵ƤǥХ		    */
/*  Input : none							    */
/*  Output: return - -ENODEV ǥХϸդʤ			    */
/*		     0	     ǥХϸդä			    */
/*--------------------------------------------------------------------------*/
static int get_device(void)
{
	struct pci_dev* pdev = NULL;

	unsigned int  apic_addr;	// APIC21 Base Address
	unsigned int  io_addr;		// I/O Base Address
	unsigned char irq;		// Interrupt Request

	int i, bsn = 0;

	for (i = 0; i < APCI8131_MAX_DEVICES; i++)
	{
		// PCI ǥХ
		pdev = pci_find_device(APCI8131_VENDOR_ID,
				       APCI8131_DEVICE_ID, pdev);
		if (pdev == NULL)
		    break;

		// APIC21 ١ɥ쥹
		apic_addr = get_device_base_addr(pdev->base_address[0]);
		//  I/O ١ɥ쥹
		io_addr	  = get_device_base_addr(pdev->base_address[1]);
		// IRQ μ
		irq	  = pdev->irq;

		// BSN μ
		bsn = get_bsn(io_addr);

		PDEBUG("apic_addr: %X\n", apic_addr);
		PDEBUG("io_addr:   %X\n", io_addr);
		PDEBUG("IRQ:      %d\n", irq);
		PDEBUG("BSN:      %d\n", bsn);

		// ΰ֤˥ǡꤵƤʤϿʤ
		if (devctl[bsn].apic_addr != 0x0000)
		{
			PWARNING("BSN not unique %d\n", bsn);
			continue;
		}

		// BSN Ʊ֤˥ǡ
		devctl[bsn].apic_addr = apic_addr;  // Apci21 Base Address save
		devctl[bsn].io_addr   = io_addr;    // I/O    Base Address save
		devctl[bsn].irq	      = irq;	    // Interrupt Request save
	}

	if (i == 0)
	{
		PWARNING("Not found Device(s)\n");
		return -ENODEV;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_base_addr - ѲǽʥǥХΥ١ɥ쥹	    */
/*  Input : base_addr - ١ɥ쥹					    */
/*  Output: return - Խ١ɥ쥹				    */
/*--------------------------------------------------------------------------*/
static unsigned int get_device_base_addr(unsigned int base_addr)
{
	// I/O ١ɥ쥹쥸
	if ((base_addr & PCI_BASE_ADDRESS_SPACE) == PCI_BASE_ADDRESS_SPACE_IO)
		return base_addr & PCI_BASE_ADDRESS_IO_MASK;

	// Memory ١ɥ쥹쥸
	return base_addr & PCI_BASE_ADDRESS_MEM_MASK;
}

/*--------------------------------------------------------------------------*/
/*  get_bsn - Board Select Number 					    */
/*  Input : io_addr - IO ݡȥɥ쥹					    */
/*  Output: return - BSN						    */
/*--------------------------------------------------------------------------*/

// aPCI-8131 BSN Ѵơ֥
static unsigned char bsn_tbl[16] = { 0x0, 0x8, 0x4, 0xc, 0x2, 0xa, 0x6, 0xe,
				     0x1, 0x9, 0x5, 0xd, 0x3, 0xb, 0x7, 0xf };

static int get_bsn(unsigned int io_addr)
{
	unsigned char bsn;

	// BSN ͼ
	bsn = APCI8131_GET_BSN(inb(io_addr + APCI8131_BSN_PORT));

	// aPCI-8131BSNѴ
	return (int)bsn_tbl[bsn];
}

/*--------------------------------------------------------------------------*/
/*  regist_char_driver - ᥸㡼ֹưŪƤȥɥ饤ФϿ	    */
/*  Input : none							    */
/*  Output: return - register_chrdev 				    */
/*		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int regist_char_driver(void)
{
	int result;

	PDEBUG("devmajor=%d\n", devmajor);

	if (devmajor == 0)
	{
		// ᥸㡼ֹ椬ꤵƤʤ
		// ưŪ᥸㡼ֹμ
		result = register_chrdev(0, APCI8131_NAME, NULL);
		if (result < 0)
		{
			PDEBUG("get major device number faild.\n");
			return result;
		}

		uapci8131major = result;
	}
	else
	{
		// ᥸㡼ֹ椬ꤵƤ
		uapci8131major = devmajor;
	}

	PDEBUG("major number=%u\n", uapci8131major);

	// 饯ǥХɥ饤ФϿ
	result = register_chrdev(uapci8131major, APCI8131_NAME, &apci8131_fops);
	if (result < 0)
	{
		PDEBUG("register_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  unregist_char_driver - ɥ饤ФϿä			    */
/*  Input : none							    */
/*  Output: return - unregister_chrdev 				    */
/*		     0	Ͽ					    */
/*--------------------------------------------------------------------------*/
static int unregist_char_driver(void)
{
	int result;

	// 饯ǥХɥ饤ФϿ
	result = unregister_chrdev(uapci8131major, APCI8131_NAME);
	if (result < 0)
	{
		PDEBUG("unregister_chrdev faild.\n");
		return result;
	}

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  get_device_state - ǥХξ֤				    */
/*  Input : ref	  - Ȱ						    */
/*  Output: return - -ENODEV ΰ֤˥ǥХϤʤ			    */
/*		     -EBUSY  					    */
/*		     0	     Ѳǽ 					    */
/*  Notes : ref >= 0 - ref <= 15 ޤǤʤ				    */
/*--------------------------------------------------------------------------*/
static int get_device_state(int ref)
{
	if (devctl[ref].apic_addr == 0x0000)
		return -ENODEV;

	if (devctl[ref].f_version != 0)
		return -EBUSY;

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  setup_hw_isr - ߥ롼				    */
/*  Input : ref	  - Ȱ						    */
/*  Output: return - ߥϥɥϿ				    */
/*--------------------------------------------------------------------------*/
static int setup_hw_isr(int ref)
{
	unsigned long flags = SA_INTERRUPT | SA_SHIRQ;	// ® | ͭ
	int result;

	// ߶ػ
	apic_set_eicr(ref, APIC21_EICR_DIS);

	// װ
	apic_set_icr(ref, APIC21_ICR_ALL);

	// ߥϥɥ
	result = request_irq(devctl[ref].irq, apci8131_isr, flags, APCI8131_NAME, (void*)(unsigned)ref);
	if (result == 0)
		return 0;

	PDEBUG("request irq faild %d\n", devctl[ref].irq);

	return result;
}

/*--------------------------------------------------------------------------*/
/*  release_hw_isr - ߥ롼β				    */
/*  Input : ref - Ȱ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static void release_hw_isr(int ref)
{
	// ߥϥɥ
	free_irq(devctl[ref].irq, (void*)(unsigned)ref);
}

/*--------------------------------------------------------------------------*/
/*  verify_arg - ǥХͭΰ				    */
/*  Input : cmd - ioctl 楳ޥ					    */
/*	    arg - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     verify_area 				    */
/*		     0	     Ѳǽ 					    */
/*--------------------------------------------------------------------------*/
static int verify_arg(unsigned int cmd, unsigned long arg)
{
	int size = _IOC_SIZE(cmd);

	/* don't even decode wrong cmds: better returning  EINVAL than EFAULT */
	if (_IOC_TYPE(cmd) != APCI8131_IOC_MAGIC)
		return -EINVAL;

	if (_IOC_NR(cmd) > APCI8131_IOC_MAXNR)
		return -EINVAL;

	/*
	 * the type is a bitmask, and VERIFY_WRITE catches R/W
	 * transfers. Note that the type is user-oriented, while
	 * verify_area is kernel-oriented, so the concept of "read" and
	 * "write" is reversed
	 */
	if (_IOC_TYPE(cmd) & _IOC_READ)
		return verify_area(VERIFY_WRITE, (void*)arg, size);

	if (_IOC_TYPE(cmd) & _IOC_WRITE)
		return verify_area(VERIFY_READ, (void*)arg, size);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_ioc_inp - ϥݡȤǡ			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_ioc_inp(struct file* file, unsigned long arg)
{
	int ref;
	apci8131_io data;

	PDEBUG("In Port\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	copy_from_user(&data, (void*)arg, sizeof(data));

	// ݡȳǧ
	if (!is_port(data.port))
		return -EINVAL;

	data.value = inb(devctl[ref].io_addr + data.port);

	// 桼֤إǡ
	copy_to_user((void*)arg, &data, sizeof(data));

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_ioc_outp - ϥݡȤإǡ			    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_ioc_outp(struct file* file, unsigned long arg)
{
	int ref;
	apci8131_io data;

	PDEBUG("Out Port\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	copy_from_user(&data, (void*)arg, sizeof(data));

	// ݡȳǧ
	if (!is_port(data.port))
		return -EINVAL;

	outb(data.value, devctl[ref].io_addr + data.port);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_ioc_get_ifr - װ				    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_ioc_get_ifr(struct file* file, unsigned long arg)
{
	int ref;

	PDEBUG("Get Interrupt Flag\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤إǡ
	put_user(devctl[ref].ifr, (unsigned char*)arg);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_ioc_get_icr - ߾				    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_ioc_get_icr(struct file* file, unsigned long arg)
{
	int ref;
	unsigned short eicr;

	PDEBUG("Set Interrupt Factor\n");

	// ǡλȰ
	ref = (int)file->private_data;

	eicr = apic_get_eicr(ref);

	// 桼֤إǡ
	put_user(eicr, (unsigned short*)arg);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_ioc_set_icr - ߾				    */
/*  Input : file - ͥ file ¤ΤΥݥ			    */
/*	    arg	 - 桼֤ؤΥɥ쥹				    */
/*  Output: return - -EINVAL ̵ʰ					    */
/*		     0	     ｪλ					    */
/*--------------------------------------------------------------------------*/
static int apci8131_ioc_set_icr(struct file* file, unsigned long arg)
{
	int ref;
	unsigned short eicr;

	PDEBUG("Set Interrupt Factor\n");

	// ǡλȰ
	ref = (int)file->private_data;

	// 桼֤ǡ
	get_user(eicr, (unsigned short*)arg);

	apic_set_eicr(ref, eicr);

	return 0;
}

/*--------------------------------------------------------------------------*/
/*  apci8131_isr - ߥӥ롼				    */
/*--------------------------------------------------------------------------*/
static void apci8131_isr(int irq, void* dev_id, struct pt_regs* regs)
{
	unsigned bsn = (unsigned)dev_id;
	unsigned char ifr;

//	PDEBUG("Interrupt bsn=%d\n", bsn);

	// ߸
	if (!(bsn >= 0 && bsn < APCI8131_MAX_DEVICES))	// ϰ
		return;

	if (devctl[bsn].f_version == 0)		// ǥХƤʤ
		return;

	ifr = apic_get_ifr((int)bsn);
	if (!(ifr & APIC21_IFR_MSK))		// ȤγߤǤϤʤ
		return;

	PDEBUG("Interrupt ifr=%02X\n", ifr);

	devctl[bsn].ifr = ifr;			// װ¸

	// װβ
	apic_set_icr((int)bsn, ifr);

	devctl[bsn].int_occurs = APCI8131_INT_OCCURS;
	wake_up_interruptible(&devctl[bsn].wq);

	PDEBUG("Wake up Interrupt\n");
}

/*--------------------------------------------------------------------------*/
/*  is_port - ݡȤ					    */
/*  Input : port - Ĵ٤ݡ					    */
/*  Output: return - 0 ̵						    */
/*		     1 						    */
/*--------------------------------------------------------------------------*/
static int is_port(unsigned int port)
{
	if (port >= APCI8131_MAX_PORTS)
		return 0;

	return 1;
}

/*--------------------------------------------------------------------------*/
/*  apic_get_ifr - װ					    */
/*  Input : ref - Ȱ						    */
/*  Output: return - װ					    */
/*--------------------------------------------------------------------------*/
static inline unsigned char apic_get_ifr(int ref)
{
	return inb(devctl[ref].apic_addr + APIC21_IFR_PORT);
}

/*--------------------------------------------------------------------------*/
/*  apic_set_icr - װꥢ					    */
/*  Input : ref	 - Ȱ						    */
/*	    icr	 - ꥢ						    */
/*  Output: none							    */
/*--------------------------------------------------------------------------*/
static inline void apic_set_icr(int ref, unsigned char icr)
{
	outb(icr, devctl[ref].apic_addr + APIC21_ICR_PORT);
}

/*--------------------------------------------------------------------------*/
/*  apic_get_eicr - ߾					    */
/*  Input : ref - Ȱ						    */
/*  Output: return - ߾					    */
/*--------------------------------------------------------------------------*/
static inline unsigned short apic_get_eicr(int ref)
{
	return inw(devctl[ref].apic_addr + APCI21_EICR_PORT);
}

/*--------------------------------------------------------------------------*/
/*  apic_set_eicr - ߾					    */
/*  Input : ref	 - Ȱ						    */
/*	    eicr - ߾						    */
/*--------------------------------------------------------------------------*/
static inline void apic_set_eicr(int ref, unsigned short eicr)
{
	outw(eicr, devctl[ref].apic_addr + APCI21_EICR_PORT);
}
